#include <lektor/common.h>
#include <lektor/logfile.h>

/* Abort function */

EXIT_FUNCTION
___lkt_abort(void)
{
    LOG_ERROR("ABORT", "Got aborted, I will provoque a segv to try to display the backtrace...");
    int *segv = (int *)-1;
    *segv     = 1;
    exit(EXIT_FAILURE);
}

/* Assert function */

PRIVATE_FUNCTION void
___no_assert(const char UNUSED *f, int UNUSED l, const char UNUSED *fn, const char UNUSED *m)
{
}

PRIVATE_FUNCTION EXIT_FUNCTION
___yes_assert(const char *file, int line, const char *func, const char *msg)
{
    LOG_ERROR("ASSERT", "Assertion failed: %s:%s:%d: %s", file, func, line, msg);
    LOG_ERROR("ASSERT", "I will provoque a segv to try to display the backtrace...");
    ___lkt_abort();
}

void (*___lkt_assert)(const char *file, int line, const char *func,
                      const char *msg) = ___yes_assert;

CONSTRUCTOR_FUNCTION
__set_assert(void)
{
    const char *no_assert = getenv(LKT_ENV_UNSET_ASSERT);
    if (NULL != no_assert && strlen(no_assert) > 0)
        ___lkt_assert = ___no_assert;
}

/* Log functions */

static bool ___log_level_verbose = false; /* Set whether the log level is verbose or contracted */
static LOG_LEVEL ___log_level    = ___LAST_UNUSED_LOG_LEVEL; /* All by default */
extern struct lkt_logfile *___log_logfile;                   /* Use the logfile globaly */

void
lkt_set_log_level_verbose(bool flag)
{
    ___log_level_verbose = flag;
}

void
lkt_set_log_level(LOG_LEVEL lvl)
{
    lkt_assert(lvl != ___LAST_UNUSED_LOG_LEVEL);
    if (lvl <= LOG_LEVEL_ERROR)
        ___log_level = LOG_LEVEL_ERROR;
    else if (lvl >= LOG_LEVEL_DEBUG)
        ___log_level = LOG_LEVEL_DEBUG;
    else
        ___log_level = lvl;
}

void
lkt_set_log_logfile(struct lkt_logfile *logfile)
{
    ___log_logfile = logfile;
}

LOG_LEVEL
lkt_get_log_level(void) { return ___log_level; }

void
___lkt_log(LOG_LEVEL level, const char *section, const char *func, const char *file,
           uint64_t line_number, const char *format, ...)
{
    lkt_assert(level != ___LAST_UNUSED_LOG_LEVEL);

    char line[LKT_MESSAGE_MAX];
    va_list ap;
    int hour = 0, min = 0, sec = 0;
    time_t rawtime     = time(NULL);
    struct tm *timeval = localtime(&rawtime);
    if (timeval != NULL) {
        hour = timeval->tm_hour;
        min  = timeval->tm_min;
        sec  = timeval->tm_sec;
    }

    size_t skip_dir = 0;
    do {
        file += skip_dir + 1;
        skip_dir = strcspn(file, "/");
    } while (file[skip_dir] && file[skip_dir + 1]);

    if (level > ___log_level)
        return;

    const char *const c_level = (level == LOG_LEVEL_ERROR) && ___log_level_verbose      ? "ERROR"
                                : (level == LOG_LEVEL_WARN) && ___log_level_verbose     ? "WARN "
                                : (level == LOG_LEVEL_INFO) && ___log_level_verbose     ? "INFO "
                                : (level == LOG_LEVEL_DEBUG) && ___log_level_verbose    ? "DEBUG"
                                : (level == LOG_LEVEL_ERROR) && (!___log_level_verbose) ? "!"
                                : (level == LOG_LEVEL_WARN) && (!___log_level_verbose)  ? "*"
                                : (level == LOG_LEVEL_INFO) && (!___log_level_verbose)  ? "."
                                : (level == LOG_LEVEL_DEBUG) && (!___log_level_verbose) ? " "
                                                                                        : " ";
    ssize_t count =
        safe_snprintf(line, LKT_MESSAGE_MAX, " %s [%02d:%02d:%02d] %-10s %s(%s+%ld): %s\n", c_level,
                      hour, min, sec, section, func, file, line_number, format);
    /* Check for overflow */
    if (count >= LKT_MESSAGE_MAX - 1) {
        line[LKT_MESSAGE_MAX - 1] = '\0';
        line[LKT_MESSAGE_MAX - 2] = '\n';
        line[LKT_MESSAGE_MAX - 3] = '.';
        line[LKT_MESSAGE_MAX - 4] = '.';
        line[LKT_MESSAGE_MAX - 5] = '.';
    }

    /* Write the log to the stderr and the logfile if possible */
    {
        char formated_line[LKT_MESSAGE_MAX];
        va_start(ap, format);
        safe_vsnprintf(formated_line, LKT_MESSAGE_MAX, line, ap);
        write(2, formated_line, strlen(formated_line));
        if (___log_logfile != NULL)
            lkt_logfile_write(___log_logfile, formated_line);
        va_end(ap);
    }
}

void
___lkt_print(const char *section, const char *format, ...)
{
    char line[LKT_MESSAGE_MAX];
    va_list ap;
    int hour = 0, min = 0, sec = 0;
    time_t rawtime     = time(NULL);
    struct tm *timeval = localtime(&rawtime);
    if (timeval != NULL) {
        hour = timeval->tm_hour;
        min  = timeval->tm_min;
        sec  = timeval->tm_sec;
    }

    ssize_t count = safe_snprintf(line, LKT_MESSAGE_MAX, "   [%02d:%02d:%02d] %-10s %s\n", hour,
                                  min, sec, section, format);
    /* Check for overflow */
    if (count >= LKT_MESSAGE_MAX - 1) {
        line[LKT_MESSAGE_MAX - 1] = '\0';
        line[LKT_MESSAGE_MAX - 2] = '\n';
        line[LKT_MESSAGE_MAX - 3] = '.';
        line[LKT_MESSAGE_MAX - 4] = '.';
        line[LKT_MESSAGE_MAX - 5] = '.';
    }

    va_start(ap, format);
    vfprintf(stdout, line, ap);
    va_end(ap);
}

/* Not implemented */

EXIT_FUNCTION
___not_implemented(const char *func, char *file, int line)
{
    LOG_ERROR("GENERAL", "Function %s in file %s at line %d not implemented", func, file, line);
    ___lkt_abort();
}

char *
trim(char *str, size_t len, char c)
{
    char *res = (char *)str;
    char *end;

    for (; len != 0 && *res == c; ++str, --len)
        continue;

    if (*res == 0)
        return res;

    end = res + len - sizeof(char);
    len = 0;

    for (; res < end && *end == c; --end, ++len)
        continue;

    if (len > 0)
        end[1] = 0;

    return res;
}

int
is_utf8(const char *string)
{
    if (!string)
        return 1;

    const unsigned char *bytes = (const unsigned char *)string;
    while (*bytes) {
        /* ASCII */
        if (bytes[0] == 0x09 || bytes[0] == 0x0A || bytes[0] == 0x0D ||
            (0x20 <= bytes[0] && bytes[0] <= 0x7E)) {
            bytes += 1;
            continue;
        }

        /* non-overlong 2-byte */
        if ((0xC2 <= bytes[0] && bytes[0] <= 0xDF) && (0x80 <= bytes[1] && bytes[1] <= 0xBF)) {
            bytes += 2;
            continue;
        }

        if ((bytes[0] == 0xE0 && (0xA0 <= bytes[1] && bytes[1] <= 0xBF) &&
             (0x80 <= bytes[2] && bytes[2] <= 0xBF)) || /* excluding overlongs */
            (((0xE1 <= bytes[0] && bytes[0] <= 0xEC) || bytes[0] == 0xEE || bytes[0] == 0xEF) &&
             (0x80 <= bytes[1] && bytes[1] <= 0xBF) &&
             (0x80 <= bytes[2] && bytes[2] <= 0xBF)) || /* straight 3-byte */
            (bytes[0] == 0xED && (0x80 <= bytes[1] && bytes[1] <= 0x9F) &&
             (0x80 <= bytes[2] && bytes[2] <= 0xBF))) { /* excluding surrogates */
            bytes += 3;
            continue;
        }

        if ((bytes[0] == 0xF0 && /* planes 1-3 */
             (0x90 <= bytes[1] && bytes[1] <= 0xBF) && (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
             (0x80 <= bytes[3] && bytes[3] <= 0xBF)) ||
            ((0xF1 <= bytes[0] && bytes[0] <= 0xF3) && (0x80 <= bytes[1] && bytes[1] <= 0xBF) &&
             (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
             (0x80 <= bytes[3] && bytes[3] <= 0xBF)) || /* planes 4-15 */
            (bytes[0] == 0xF4 && (0x80 <= bytes[1] && bytes[1] <= 0x8F) &&
             (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
             (0x80 <= bytes[3] && bytes[3] <= 0xBF))) { /* plane 16 */
            bytes += 4;
            continue;
        }

        return 2;
    }

    return 0;
}

void *
safe_malloc(size_t size)
{
    void *tmp = malloc(size);
    if (NULL == tmp)
        LOG_FATAL("Out of memory, can't alloc '%ld' Bytes", size);
    memset(tmp, 0, size);
    return tmp;
}

void *
safe_realloc(void *ptr, unsigned long size)
{
    void *tmp = realloc(ptr, size);
    if (NULL == tmp)
        LOG_FATAL("Out of memory, can't realloc '%ld' bytes for ptr %p", size, ptr);
    return tmp;
}

void
safe_free(void **ptr)
{
    void *pptr = *ptr;
    if (pptr) {
        free(pptr);
        *ptr = NULL;
    }
}

#define ___internal_vsnprintf(dest, len, format, ap)  \
    int ret       = vsnprintf(dest, len, format, ap); \
    dest[len - 1] = '\0';                             \
    va_end(ap);                                       \
    return ret;

int
safe_snprintf(char *dest, const size_t max_len, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    ___internal_vsnprintf(dest, max_len, format, ap);
}

int
safe_vsnprintf(char *dest, const size_t max_len, const char *format, va_list src)
{
    va_list ap;
    va_copy(ap, src);
    ___internal_vsnprintf(dest, max_len, format, ap);
}

#undef ___internal_vsnprintf

long
long_length(long integer)
{
    long count = 0;
    while (integer) {
        integer /= 10;
        ++count;
    }
    return count;
}

char *
safe_strncpy(char *dest, const char *src, size_t n)
{
    /* Edge cases */
    if (src == NULL && n > 0) {
        dest[0] = '\0';
        return dest;
    } else if (n == 0) {
        return dest;
    }

    /* Regular case */
    char *ret   = strncpy(dest, src, n);
    dest[n - 1] = '\0';
    return ret;
}

int
iter_string(const char *str_list, const char *sep, char *dest, const size_t dest_len, size_t *save)
{
    size_t len = strcspn(&str_list[*save], sep);

    if (len == 0)
        return 1;
    if (len >= dest_len)
        return -1;

    strncpy(dest, &str_list[*save], len);
    dest[len] = '\0';
    *save += len + 1;
    return 0;
}
