#include <lektor/stack.h>
#include <lektor/common.h>

void
stack_new(struct stack *ret)
{
    lkt_assert(ret);

    struct stack _ret = {
        .contents = LKT_ALLOC_ARRAY(volatile void *, LKT_DEFAULT_LIST_SIZE),
        .size     = LKT_DEFAULT_LIST_SIZE,
        .len      = 0,
    };

    *ret = _ret;
}

void
stack_free(struct stack *lst)
{
    if (!lst)
        return;

    if (lst->contents)
        free(lst->contents);

    lst->size = 0;
    lst->len  = 0;
}

int
stack_empty(struct stack *lst)
{
    if (!lst || !lst->contents || !lst->len)
        return 1;
    return 0;
}

void
stack_push(struct stack *lst, void *item)
{
    lkt_assert(lst);

    if (lst->size == lst->len) {
        volatile void **new = realloc(lst->contents, lst->size * 2 * sizeof(void *));

        if (NULL == new) {
            LOG_ERROR("GENERAL", "Out of memory");
            abort();
        }

        lst->contents = new;
        lst->size *= 2;
    }

    lst->contents[(lst->len)++] = item;
}

int
stack_pop(struct stack *lst, void **item)
{
    if (!lst || !lst->len)
        return 1;

    *item = (void *)lst->contents[--(lst->len)];
    return 0;
}

int
stack_head(struct stack *lst, void **item)
{
    if (!lst || !lst->len)
        return 1;

    *item = (void *)lst->contents[lst->len - 1];
    return 0;
}
